from collections import Counter
from itertools import chain

class DbtModelParser():

    def __init__(self, model, dbt_model_prefix):
        self.model = model
        self.eppo_entities = []
        self.eppo_timestamp = []
        self.eppo_facts = []
        self.eppo_properties = []
        self.dbt_model_prefix = dbt_model_prefix

    def _parse_column_entities(self, column, column_entities):

        for column_entity in column_entities:
            column_entity_split = column_entity.split(':')
            # expected format: eppo_entity:<entity name>
            EXPECTED_SPLIT_LENGTH = 2
            assert len(column_entity_split) == EXPECTED_SPLIT_LENGTH, \
                f'Invalid entity tag {column_entity} in model {self.model["name"]}'
            self.eppo_entities.append({
                "entity_name": column_entity_split[1],
                "column": column['name']
            })     


    def _parse_one_column(self, column):

        tags = column.get('tags')

        if tags:

            column_entities = [t for t in tags if 'eppo_entity:' in t]

            if column_entities:
                self._parse_column_entities(column, column_entities)

            if 'eppo_timestamp' in tags:
                self.eppo_timestamp = column["name"]

            if 'eppo_fact' in tags:
                self.eppo_facts.append({
                    "name": column["name"],
                    "column": column["name"],
                    "description": column.get("description", "")
                })
            
            if 'eppo_property' in tags:
                self.eppo_properties.append({
                    "name": column["name"],
                    "column": column["name"],
                    "description": column.get("description", "")
                })

    
    def parse_columns(self):

        for column in self.model.get('columns'):
            self._parse_one_column(column)

    
    def validate(self):

        validation_errors = []

        # make sure there is at least one entity and exactly one timestamp
        if len(self.eppo_timestamp) == 0:
            validation_errors.append(
                'Exactly 1 column must be have tag "eppo_timestamp"'
            )
        
        if len(self.eppo_entities) == 0:
            validation_errors.append(
                'At least 1 column must have tag "eppo_entity:<entity_name>"'
            )
        
        # check that no columns are tagged to multiple Eppo fields (skip if there
        # is already a validation error)

        if len(validation_errors) == 0:
            # parse names from column list
            entity_names = [e['column'] for e in self.eppo_entities]
            timestamp_names = [self.eppo_timestamp]
            fact_names = [e['column'] for e in self.eppo_facts]
            property_names = [e['column'] for e in self.eppo_properties]

            all_names = Counter(chain.from_iterable(
                [entity_names, timestamp_names, fact_names, property_names]
            ))

            overlapping_columns = [item for item, count in all_names.items() if count > 1]

            if len(overlapping_columns):
                validation_errors.append(
                    f'The following columns had tags to multiple Eppo fields: {", ".join(overlapping_columns)}'
                )
            
        
        if len(validation_errors):
            raise ValueError(
                f'One or more errors parsing model: {self.model["name"]}: {", ".join(validation_errors)}'
            )

    
    def format(self):

        entity_clause = '\n  , '.join([e['column'] for e in self.eppo_entities])
        fact_clause = '\n  , '.join([f['column'] for f in self.eppo_facts])
        property_clause = '\n  , '.join([p['column'] for p in self.eppo_properties])

        formatted_sql = \
         f"""
        select
            {self.eppo_timestamp}
            , {entity_clause}
            , {fact_clause}
            , {property_clause}
        from 
            {self.dbt_model_prefix}.{self.model['name']}
        """

        self.eppo_fact_source = {
            "name": self.model["name"],
            "sql": formatted_sql,
            "timestamp_column": self.eppo_timestamp,
            "entities": self.eppo_entities,
            "facts": self.eppo_facts,
            "properties": self.eppo_properties
        }
    
    def build(self):
        if isinstance(self.model, dict):
            model_tags = self.model.get('tags', [])
            if 'eppo_fact_source' in model_tags:
                self.parse_columns()
                self.validate()
                self.format()
                return self.eppo_fact_source
        else:
            raise ValueError(f"Expected model to be a dictionary, got model = {self.model}")



from collections import Counter

advanced_aggregation_parameters = [
    'retention_threshold_days',
    'conversion_threshold_days',
    'threshold_metric_settings'
]

winsorization_parameters = [
    'winsorization_lower_percentile',
    'winsorization_upper_percentile'
]

timeframe_parameters = [
    'aggregation_timeframe_start_value',
    'aggregation_timeframe_end_value',
    'aggregation_timeframe_unit'
]


def check_for_duplicated_names(payload, names, object_name):
    element_counts = Counter(names)
    duplicate_elements = [i for i, count in element_counts.items() if count > 1]
    if duplicate_elements:
        payload.validation_errors.append(
            object_name + ' names are not unique: ' + ', '.join(duplicate_elements)
        )


def unique_names(payload):
    fact_source_names = []
    fact_names = []
    fact_property_names = []

    for fact_source in payload.fact_sources:
        fact_source_names.append(fact_source['name'])
        fact_names.extend([f['name'] for f in fact_source['facts']])
        if ('properties' in fact_source):
            fact_property_names.extend(
                [f['name'] for f in fact_source['properties']]
            )

    metric_names = [m['name'] for m in payload.metrics]

    check_for_duplicated_names(payload, fact_source_names, 'Fact source')
    check_for_duplicated_names(payload, fact_names, 'Fact')
    # TODO: check for distinct names within a given fact source
    # check_for_duplicated_names(payload, fact_property_names, 'Fact property')
    check_for_duplicated_names(payload, metric_names, 'Metric')

    return True


def valid_fact_references(payload):
    fact_references = set()
    for metric in payload.metrics:
        if 'numerator' in metric:
            fact_references.add(metric['numerator']['fact_name'])
        if 'denominator' in metric:
            fact_references.add(metric['denominator']['fact_name'])
        if 'percentile' in metric:
            fact_references.add(metric['percentile']['fact_name'])

    fact_names = set()
    for fact_source in payload.fact_sources:
        for fact in fact_source['facts']:
            fact_names.add(fact['name'])

    if fact_references.issubset(set(fact_names)) == False:
        payload.validation_errors.append(
            "Invalid fact reference(s): " +
            str(', '.join(fact_references.difference(fact_names)))
        )

def valid_experiment_computation(payload):
    for fact_source in payload.fact_sources:
        if 'properties' in fact_source:
            for property in fact_source['properties']:
                if 'include_experiment_computation' in property:
                    if not isinstance(property['include_experiment_computation'], bool):
                        payload.validation_errors.append(
                            f"Invalid include_experiment_computation value. It must be a boolean value for property: {property['name']}"
                        )

def metric_aggregation_is_valid(payload):
    for m in payload.metrics:
        if m.get('type') == 'percentile':
            percentile_error = percentile_metric_is_valid(m)
            if percentile_error:
                payload.validation_errors.append(
                    f"{m['name']} has invalid percentile configuration: {percentile_error}"
                )
            # Skip the rest of the loop iteration for percentile metrics
            # since they don't have numerator/denominator to validate
            continue

        if 'numerator' in m:
            numerator_error = aggregation_is_valid(m['numerator'])
            if numerator_error:
                payload.validation_errors.append(
                    f"{m['name']} has invalid numerator: {numerator_error}"
                )

        if 'denominator' in m:
            denominator_error = aggregation_is_valid(m['denominator'])
            if denominator_error:
                payload.validation_errors.append(
                    f"{m['name']} has invalid denominator: {denominator_error}"
                )


def valid_guardrail_cutoff_signs(payload):
    facts = dict()
    for fact_source in payload.fact_sources:
        for fact in fact_source['facts']:
            facts[fact['name']] = fact

    for m in payload.metrics:
        if m.get('type') == 'percentile':
            if is_guardrail_cutoff_exist(m):
                percentile_fact_name = m['percentile']['fact_name']
                if percentile_fact_name in facts and 'desired_change' in facts[percentile_fact_name]:
                    error = is_valid_guardrail_cutoff_sign(m, facts[percentile_fact_name])
                    if error:
                        payload.validation_errors.append(
                            f"{m['name']} is having invalid guardrail_cutoff sign: {error}"
                        )
            # Skip the rest of the loop iteration for percentile metrics
            # since they don't have numerator/denominator to validate
            continue

        if 'numerator' in m:
            numerator_fact_name = m['numerator']['fact_name']
            if is_guardrail_cutoff_exist(m) and numerator_fact_name in facts and 'desired_change' in facts[numerator_fact_name]:
                error = is_valid_guardrail_cutoff_sign(m, facts[numerator_fact_name])
                if error:
                    payload.validation_errors.append(
                        f"{m['name']} is having invalid guardrail_cutoff sign: {error}"
                    )


def is_valid_guardrail_cutoff_sign(metric, numerator_fact):
    if numerator_fact['desired_change'] == 'decrease' and metric['guardrail_cutoff'] < 0:
        return f"guardrail_cutoff value should be positive"
    elif numerator_fact['desired_change'] == 'increase' and metric['guardrail_cutoff'] > 0:
        return f"guardrail_cutoff value should be negative"


def is_guardrail_cutoff_exist(metric):
    return 'is_guardrail' in metric and metric['is_guardrail'] and 'guardrail_cutoff' in metric


def distinct_advanced_aggregation_parameter_set(
        aggregation,
        operation,
        aggregation_parameter,
        error_message
):
    if aggregation['operation'] == operation:
        matched = [p for p in advanced_aggregation_parameters if p in aggregation]
        if len(matched) == 0:
            error_message.append(
                f'{operation} aggregation must have {aggregation_parameter} set'
            )
        elif len(matched) == 1 and matched[0] == aggregation_parameter:
            return True
        else:
            invalid_elements = [m for m in matched if m != aggregation_parameter]
            error_message.append(
                f"Invalid parameter for {operation} aggregation: {', '.join(invalid_elements)}"
            )


def aggregation_is_valid(aggregation):
    error_message = []

    if aggregation['operation'] not in ['sum', 'count', 'count_distinct', 'distinct_entity', 'threshold', 'retention',
                                        'conversion', 'last_value', 'first_value']:
        error_message.append(
            'Invalid aggregation operation: ' + aggregation['operation']
        )

    # can only winsorize sum or count metrics
    if aggregation['operation'] not in ['sum', 'count', 'last_value', 'first_value']:
        if [name for name in winsorization_parameters if name in aggregation]:
            error_message.append(
                'Cannot winsorize a metric with operation ' + aggregation['operation']
            )

    # The aggregation_timeframe_unit must be specified if timeframe parameters are set
    included_timeframe_parameters = [name for name in timeframe_parameters if name in aggregation]

    if 'aggregation_timeframe_value' in aggregation:
        error_message.append(
            'The aggregation_timeframe_value parameter has been deprecated. Please use aggregation_timeframe_end instead.'
        )

    timeframe_unit_specified = 'aggregation_timeframe_unit' in included_timeframe_parameters
    if len(included_timeframe_parameters) > 0 and not timeframe_unit_specified:
        error_message.append(
            'The aggregation_timeframe_unit must be set to use timeframe parameters.'
        )

    # only set timeframe_parameters on a some operation types
    if aggregation['operation'] in ['conversion']:
        matched = [p for p in timeframe_parameters if p in aggregation]
        if matched:
            error_message.append(
                "Cannot specify " + matched[0] + " for operation " + aggregation['operation']
            )

    # can't specify advanced aggregation parameters for simple aggregation types
    if aggregation['operation'] not in ['threshold', 'retention', 'conversion']:
        matched = [p for p in advanced_aggregation_parameters if p in aggregation]
        if matched:
            error_message.append(
                matched[0] + ' specified, but operation is ' + aggregation['operation']
            )

    # if threshold:
    if aggregation['operation'] == 'threshold':
        # TODO
        pass

    distinct_advanced_aggregation_parameter_set(
        aggregation,
        'retention',
        'retention_threshold_days',
        error_message
    )
    distinct_advanced_aggregation_parameter_set(
        aggregation,
        'conversion',
        'conversion_threshold_days',
        error_message
    )
    distinct_advanced_aggregation_parameter_set(
        aggregation,
        'threshold',
        'threshold_metric_settings',
        error_message
    )

    if error_message:
        return '\n'.join(error_message)
    else:
        return None

def percentile_metric_is_valid(metric):
    error_message = []

    # Check for required percentile field
    if 'percentile' not in metric:
        error_message.append("Missing 'percentile' field for percentile metric")
        return '\n'.join(error_message)

    percentile = metric['percentile']

    # Check for required fact_name
    if 'fact_name' not in percentile:
        error_message.append("Missing 'fact_name' in percentile configuration")

    # Check for required percentile_value
    if 'percentile_value' not in percentile:
        error_message.append("Missing 'percentile_value' in percentile configuration")
    elif not isinstance(percentile['percentile_value'], (int, float)):
        error_message.append("'percentile_value' must be a number")
    elif percentile['percentile_value'] < 0 or percentile['percentile_value'] > 1:
        error_message.append("'percentile_value' must be between 0 and 1")

    if error_message:
        return '\n'.join(error_message)
    else:
        return None



import pytest

from eppo_metrics_sync.validation import unique_names, valid_fact_references, aggregation_is_valid
from eppo_metrics_sync.eppo_metrics_sync import EppoMetricsSync

test_yaml_dir = "tests/yaml/dbt/"


def test_invalid_entity_tag():

    eppo_metrics_sync = EppoMetricsSync(
        directory = None, 
        schema_type = 'dbt-model',
        dbt_model_prefix = 'foo'
    )
    with pytest.raises(
        AssertionError, 
        match = "Invalid entity tag eppo_entity:anonymous_user:foo in model revenue"
        ):
        eppo_metrics_sync.load_dbt_yaml(path = test_yaml_dir + "invalid/invalid_entity_tag.yml")


def test_missing_entity():

    eppo_metrics_sync = EppoMetricsSync(
        directory = None, 
        schema_type = 'dbt-model',
        dbt_model_prefix = 'foo'
    )
    with pytest.raises(
        ValueError, 
        match = 'At least 1 column must have tag "eppo_entity:<entity_name>"'
        ):
        eppo_metrics_sync.load_dbt_yaml(path = test_yaml_dir + "invalid/missing_entity.yml")
    

def test_missing_timestamp():

    eppo_metrics_sync = EppoMetricsSync(
        directory = None, 
        schema_type = 'dbt-model',
        dbt_model_prefix = 'foo'
    )
    with pytest.raises(
        ValueError, 
        match = 'Exactly 1 column must be have tag "eppo_timestamp"'
        ):
        eppo_metrics_sync.load_dbt_yaml(path = test_yaml_dir + "invalid/missing_timestamp.yml")

  
def test_overlapping_tags():

    eppo_metrics_sync = EppoMetricsSync(
        directory = None, 
        schema_type = 'dbt-model',
        dbt_model_prefix = 'foo'
    )
    with pytest.raises(
        ValueError, 
        match = 'The following columns had tags to multiple Eppo fields: gross_revenue'
        ):
        eppo_metrics_sync.load_dbt_yaml(path = test_yaml_dir + "invalid/overlapping_tags.yml")


def test_model_is_not_a_dictionary():

    eppo_metrics_sync = EppoMetricsSync(
        directory = None, 
        schema_type = 'dbt-model',
        dbt_model_prefix = 'foo'
    )
    with pytest.raises(
        ValueError, 
        match = 'Expected model to be a dictionary, got model = r'
        ):
        eppo_metrics_sync.load_dbt_yaml(path = test_yaml_dir + "invalid/model_is_not_a_dictionary.yml")

# test that the package handles yml without 'models' member gracefully
def test_no_model_tag():

    eppo_metrics_sync = EppoMetricsSync(
        directory = None, 
        schema_type = 'dbt-model',
        dbt_model_prefix = 'foo'
    )
    eppo_metrics_sync.load_dbt_yaml(path = test_yaml_dir + "valid/no_model_property.yml")


# test that the package handles dbt models without eppo tags gracefully
def test_no_model_tag():

    eppo_metrics_sync = EppoMetricsSync(
        directory = None, 
        schema_type = 'dbt-model',
        dbt_model_prefix = 'foo'
    )
    eppo_metrics_sync.load_dbt_yaml(path = test_yaml_dir + "valid/no_dbt_tags.yml")



import subprocess
import pytest

pytest_plugins = ["pytester"]

@pytest.fixture
def run_cli():
    def runner(args):
        result = subprocess.run(['python3', '-m', 'eppo_metrics_sync', *args], 
                                stdout=subprocess.PIPE, 
                                stderr=subprocess.PIPE, 
                                text=True)
        return result
    return runner

"""def test_cli_with_valid_directory(run_cli):
    result = run_cli(['tests/yaml/valid'])
    assert result.returncode == 0
"""

def test_cli_dbt_valid_directory(run_cli):
    result = run_cli([
        '--dryrun',
        '--schema=dbt-model', 
        '--dbt-model-prefix=foo',
        'tests/yaml/dbt/valid',
        ])



import pytest

from eppo_metrics_sync.validation import unique_names, valid_fact_references, aggregation_is_valid
from eppo_metrics_sync.eppo_metrics_sync import EppoMetricsSync


def test_dbt():
    eppo_metric_sync = EppoMetricsSync(
        'tests/yaml/dbt/valid', 
        schema_type='dbt-model', 
        dbt_model_prefix='test_db.test_schema'
    )
    eppo_metric_sync.read_yaml_files()
    eppo_metric_sync.validate()



import pytest
import re

from eppo_metrics_sync.validation import unique_names, valid_fact_references, aggregation_is_valid
from eppo_metrics_sync.eppo_metrics_sync import EppoMetricsSync

# If we use context.py we can do something like this instead
# from .context import eppo_metric_sync
# from .context import validation


test_yaml_dir = "tests/yaml/invalid"


def test_unique_fact_source_names():
    eppo_metrics_sync = EppoMetricsSync(directory=None)
    eppo_metrics_sync.load_eppo_yaml(
        path=test_yaml_dir + "/duplicated_fact_source_names.yaml")

    with pytest.raises(ValueError, match="Fact source names are not unique: upgrades_table"):
        eppo_metrics_sync.validate()


def test_unique_metric_names():
    eppo_metrics_sync = EppoMetricsSync(directory=None)
    eppo_metrics_sync.load_eppo_yaml(
        path=test_yaml_dir + "/duplicated_metric_names.yaml")

    with pytest.raises(ValueError, match="Metric names are not unique: Total Upgrades to Paid Plan"):
        eppo_metrics_sync.validate()


def test_unique_fact_names():
    eppo_metrics_sync = EppoMetricsSync(directory=None)
    eppo_metrics_sync.load_eppo_yaml(
        path=test_yaml_dir + "/duplicated_fact_names.yaml")

    with pytest.raises(ValueError, match="Fact names are not unique: upgrades"):
        eppo_metrics_sync.validate()


def test_valid_guardrail_cutoff_signs():
    eppo_metrics_sync = EppoMetricsSync(directory=None)
    eppo_metrics_sync.load_eppo_yaml(
        path=test_yaml_dir + "/invalid_guardrail_cutoff_sign.yaml")

    with pytest.raises(ValueError, match="Total Upgrades to Paid Plan is having invalid guardrail_cutoff sign: "
                                         "guardrail_cutoff value should be negative"):
        eppo_metrics_sync.validate()


"""def test_unique_fact_property_names():
    eppo_metrics_sync = EppoMetricsSync(directory = None)
    eppo_metrics_sync.load_eppo_yaml(
        path = test_yaml_dir + "/duplicated_fact_property_names.yaml")
    with pytest.raises(ValueError, match = "Fact property names are not unique: device"):
        eppo_metrics_sync.validate()"""


def test_invalid_fact_reference():
    eppo_metrics_sync = EppoMetricsSync(directory=None)
    eppo_metrics_sync.load_eppo_yaml(
        path=test_yaml_dir + "/invalid_fact_reference.yaml")
    with pytest.raises(ValueError, match=re.escape("Invalid fact reference(s): revenue")):
        eppo_metrics_sync.validate()

def test_invalid_experiment_computation():
    eppo_metrics_sync = EppoMetricsSync(directory=None)
    eppo_metrics_sync.load_eppo_yaml(
        path=test_yaml_dir + "/invalid_experiment_computation.yaml")
    with pytest.raises(ValueError, match=re.escape("Invalid include_experiment_computation value. It must be a boolean value for property: Traffic Source")):
        eppo_metrics_sync.validate()

def test_invalid_winsorization_operation():
    test_agg = {
        'operation': 'distinct_entity',
        'winsorization_lower_percentile': 0.1
    }
    res = aggregation_is_valid(test_agg)
    assert res == 'Cannot winsorize a metric with operation distinct_entity'


def test_invalid_aggregation_for_timeframe():
    test_agg = {
        'operation': 'conversion',
        'aggregation_timeframe_end_value': 1,
        'aggregation_timeframe_unit': 'days',
        'conversion_threshold_days': 1
    }

    res = aggregation_is_valid(test_agg)
    assert res == 'Cannot specify aggregation_timeframe_end_value for operation conversion'


def test_invalid_timeframe_parameters():
    test_agg = {
        'operation': 'sum',
        'aggregation_timeframe_end_value': 1
    }

    expected_error = 'The aggregation_timeframe_unit must be set to use timeframe parameters.'

    res = aggregation_is_valid(test_agg)
    assert res == expected_error


def test_invalid_aggregation_parameter():
    test_agg = {
        'operation': 'sum',
        'retention_threshold_days': 1
    }

    res = aggregation_is_valid(test_agg)
    assert res == 'retention_threshold_days specified, but operation is sum'


def test_missing_conversion_threshold():
    test_agg = {
        'operation': 'conversion'
    }

    res = aggregation_is_valid(test_agg)
    assert res == 'conversion aggregation must have conversion_threshold_days set'


def test_extra_parameter_on_retention_metric():
    test_agg = {
        'operation': 'retention',
        'conversion_threshold_days': 2
    }

    res = aggregation_is_valid(test_agg)
    assert res == 'Invalid parameter for retention aggregation: conversion_threshold_days'


def test_count_distinct():
    test_agg = {
        'operation': 'count_distinct',
        'aggregation_timeframe_start_value': 1,
        'aggregation_timeframe_end_value': 7,
        'aggregation_timeframe_unit': 'days'
    }

    res = aggregation_is_valid(test_agg)
    assert res == None

def test_last_value():
    res = aggregation_is_valid({'operation': 'last_value'})
    assert res == None

def test_first_value():
    res = aggregation_is_valid({'operation': 'first_value'})
    assert res == None

def test_valid_yaml():
    eppo_metrics_sync = EppoMetricsSync(directory=None)
    eppo_metrics_sync.load_eppo_yaml(path='tests/yaml/valid/purchases.yaml')
    eppo_metrics_sync.validate()

def test_valid_percentile_yaml():
    eppo_metrics_sync = EppoMetricsSync(directory=None)
    eppo_metrics_sync.load_eppo_yaml(path='tests/yaml/valid/percentile_test.yaml')
    eppo_metrics_sync.validate()

def test_invalid_percentile_yaml():
    eppo_metrics_sync = EppoMetricsSync(directory=None)
    eppo_metrics_sync.load_eppo_yaml(path='tests/yaml/invalid/invalid_percentile.yaml')

    with pytest.raises(ValueError) as excinfo:
        eppo_metrics_sync.validate()

    error_msg = str(excinfo.value)
    assert "Invalid App opens percentile has invalid percentile configuration: Missing 'percentile_value' in percentile configuration" in error_msg
    assert "Invalid percentile value has invalid percentile configuration: 'percentile_value' must be between 0 and 1" in error_msg

def test_percentile_metrics():
    # Test with missing percentile field
    metric = {
        'name': 'Test Percentile Metric',
        'type': 'percentile',
        'entity': 'User'
    }
    from eppo_metrics_sync.validation import percentile_metric_is_valid
    error = percentile_metric_is_valid(metric)
    assert error == "Missing 'percentile' field for percentile metric"

    # Test with missing fact_name
    metric = {
        'name': 'Test Percentile Metric',
        'type': 'percentile',
        'entity': 'User',
        'percentile': {
            'percentile_value': 0.95
        }
    }
    error = percentile_metric_is_valid(metric)
    assert error == "Missing 'fact_name' in percentile configuration"

    # Test with missing percentile_value
    metric = {
        'name': 'Test Percentile Metric',
        'type': 'percentile',
        'entity': 'User',
        'percentile': {
            'fact_name': 'App open'
        }
    }
    error = percentile_metric_is_valid(metric)
    assert error == "Missing 'percentile_value' in percentile configuration"

    # Test with invalid percentile_value type
    metric = {
        'name': 'Test Percentile Metric',
        'type': 'percentile',
        'entity': 'User',
        'percentile': {
            'fact_name': 'App open',
            'percentile_value': 'invalid'
        }
    }
    error = percentile_metric_is_valid(metric)
    assert error == "'percentile_value' must be a number"

    # Test with percentile_value out of range
    metric = {
        'name': 'Test Percentile Metric',
        'type': 'percentile',
        'entity': 'User',
        'percentile': {
            'fact_name': 'App open',
            'percentile_value': 1.5
        }
    }
    error = percentile_metric_is_valid(metric)
    assert error == "'percentile_value' must be between 0 and 1"

    # Test with valid percentile metric
    metric = {
        'name': 'Test Percentile Metric',
        'type': 'percentile',
        'entity': 'User',
        'percentile': {
            'fact_name': 'App open',
            'percentile_value': 0.95
        }
    }
    error = percentile_metric_is_valid(metric)
    assert error is None



import yaml

def load_yaml(path):
    try:
        with open(path, 'r') as file:
            content = yaml.safe_load(file)
            return content
    except yaml.YAMLError as e:
        raise ValueError(f"Error loading YAML file '{path}': {e}")
    except Exception as e:
        raise ValueError(f"Unexpected error loading file '{path}': {e}")





import subprocess
import pytest

pytest_plugins = ["pytester"]

@pytest.fixture
def run_cli():
    def runner(args):
        result = subprocess.run(['python3', '-m', 'eppo_metrics_sync', *args], 
                                stdout=subprocess.PIPE, 
                                stderr=subprocess.PIPE, 
                                text=True)
        return result
    return runner

def test_cli_dryrun_option(run_cli):
    result = run_cli(['tests/yaml/valid', '--dryrun'])
    assert result.returncode == 0

def test_cli_invalid_directory(run_cli):
    result = run_cli(['tests/yaml/invalid'])
    assert result.returncode != 0



import sys
import argparse
from eppo_metrics_sync.eppo_metrics_sync import EppoMetricsSync

if __name__ == '__main__':

    parser = argparse.ArgumentParser(
        description="Scan specified directory for Eppo yaml files and sync with Eppo"
    )
    parser.add_argument("directory", help="The directory of yaml files to process")
    parser.add_argument("--allow-upgrades", action="store_true", help="Allow existing non-certified metrics/fact sources to become certified")
    parser.add_argument("--dryrun", action="store_true", help="Run in dry run mode")
    parser.add_argument("--schema", help="One of: eppo[default], dbt-model", default='eppo')
    parser.add_argument("--sync-prefix", help="Used for testing in a shared Q/A workspace. "
                                              "Will use this as a sync tag and append all fact and metric definitions with this prefix.",
                        required=False
                        )
    parser.add_argument(
        "--dbt-model-prefix",
        help="The warehouse and schema where the dbt models live",
        default=None
    )

    args = parser.parse_args()

    eppo_metrics_sync = EppoMetricsSync(
        directory=args.directory,
        schema_type=args.schema,
        dbt_model_prefix=args.dbt_model_prefix,
        sync_prefix=args.sync_prefix,
        allow_upgrades=args.allow_upgrades
    )

    if args.dryrun:
        eppo_metrics_sync.read_yaml_files()
        eppo_metrics_sync.validate()
    else:
        eppo_metrics_sync.sync()



import json
import jsonschema
import os
import requests

from eppo_metrics_sync.validation import (
    unique_names,
    valid_fact_references,
    metric_aggregation_is_valid,
    valid_guardrail_cutoff_signs,
    valid_experiment_computation
)

from eppo_metrics_sync.dbt_model_parser import DbtModelParser
from eppo_metrics_sync.helper import load_yaml

host = os.getenv('EPPO_API_HOST', 'https://eppo.cloud')
API_ENDPOINT = f'{host}/api/v1/metrics/sync'

class EppoMetricsSync:
    def __init__(
            self,
            directory,
            schema_type='eppo',
            dbt_model_prefix=None,
            sync_prefix=None,
            allow_upgrades=False
    ):
        self.directory = directory
        self.fact_sources = []
        self.metrics = []
        self.validation_errors = []
        self.schema_type = schema_type
        self.dbt_model_prefix = dbt_model_prefix
        self.sync_prefix = sync_prefix
        self.allow_upgrades = allow_upgrades

        # temporary: ideally would pull this from Eppo API
        package_root = os.path.dirname(os.path.abspath(__file__))
        schema_path = os.path.join(package_root, 'schema', 'eppo_metric_schema.json')
        with open(schema_path) as schema_file:
            self.schema = json.load(schema_file)

    def load_eppo_yaml(self, path):
        yaml_data = load_yaml(path)
        if 'fact_sources' in yaml_data:
            self.fact_sources.extend(yaml_data['fact_sources'])
        if 'metrics' in yaml_data:
            self.metrics.extend(yaml_data['metrics'])

    def load_dbt_yaml(self, path):
        if not self.dbt_model_prefix:
            raise ValueError('Must specify dbt_model_prefix when schema_type=dbt-model')
        yaml_data = load_yaml(path)
        models = yaml_data.get('models')
        if models:
            for model in models:
                dbt_model_parser = DbtModelParser(model, self.dbt_model_prefix).build()
                if dbt_model_parser:
                    self.fact_sources.append(dbt_model_parser)

    def yaml_is_valid(self, yaml_path):
        """
        Validate a single YAML file against the schema

        """
        data = load_yaml(yaml_path)
        try:
            jsonschema.validate(data, self.schema)
            return {"passed": True}
        except jsonschema.exceptions.ValidationError as e:
            return {"passed": False, "error_message": e}

    def read_yaml_files(self):
        # Recursively scan the directory for YAML files and load valid ones
        for root, _, files in os.walk(self.directory):
            for file in files:
                if file.endswith(".yaml") or file.endswith(".yml"):

                    yaml_path = os.path.join(root, file)

                    if self.schema_type == 'eppo':
                        valid = self.yaml_is_valid(yaml_path)
                        if valid['passed']:
                            self.load_eppo_yaml(yaml_path)
                        else:
                            self.validation_errors.append(
                                f"Schema violation in {yaml_path}: \n{valid['error_message']}"
                            )

                    elif self.schema_type == 'dbt-model':
                        self.load_dbt_yaml(yaml_path)

                    else:
                        raise ValueError(f'Unexpected schema_type: {self.schema_type}')

        if len(self.fact_sources) == 0 and len(self.metrics) == 0:
            raise ValueError(
                'No valid yaml files found. ' + ', '.join(self.validation_errors)
            )

    def _add_sync_prefix(self):
        for source in self.fact_sources:
            source['name'] = f"[{self.sync_prefix}] {source['name']}"

        for metric in self.metrics:
            metric['name'] = f"[{self.sync_prefix}] {metric['name']}"

    def validate(self):

        if len(self.fact_sources) == 0 and len(self.metrics) == 0:
            raise ValueError('No fact sources or metrics found, did you call eppo_metrics.read_yaml_files()?')

        unique_names(self)
        valid_fact_references(self)
        metric_aggregation_is_valid(self)
        valid_guardrail_cutoff_signs(self)
        valid_experiment_computation(self)

        if self.validation_errors:
            error_count = len(self.validation_errors)
            error_message = f"Validation failed with {error_count} error(s): \n"
            error_message += '\n'.join(self.validation_errors)
            raise ValueError(error_message)

        return True

    def _determine_sync_tag(self):
        if self.sync_prefix is not None:
            return self.sync_prefix

        return os.getenv('EPPO_SYNC_TAG')

    def sync(self):
        self.read_yaml_files()
        if self.sync_prefix is not None:
            self._add_sync_prefix()
        self.validate()

        api_key = os.getenv('EPPO_API_KEY')
        if not api_key:
            raise Exception('EPPO_API_KEY not set in environment variables. Please set and try again')

        sync_tag = self._determine_sync_tag()
        if not sync_tag:
            raise Exception('EPPO_SYNC_TAG not set in environment variables. Please set and try again')

        headers = {"X-Eppo-Token": api_key}
        payload = {
            "sync_tag": sync_tag,
            "fact_sources": self.fact_sources,
            "metrics": self.metrics
        }

        response = requests.post(f'{API_ENDPOINT}{"?allow_upgrades=true" if self.allow_upgrades else ""}', json=payload, headers=headers)

        if response.status_code < 400:
            print('Metrics synced')
        else:
            raise Exception(f"Request failed {response.status_code}: {response.text}")

        return response
